/* Author: Wietse Venema <wietse@wzv.win.tue.nl> */

#define _PATH_DEFAULT	"/etc/default/login"

extern char *default_console;
extern char *default_altsh;
extern char *default_passreq;
extern char *default_timezone;
extern char *default_hz;
extern char *default_path;
extern char *default_supath;
extern char *default_ulimit;
extern char *default_timeout;
extern char *default_umask;
